# -*- coding: utf-8 -*-
import urllib.parse

import humanize


class Item:
    def __init__(self, name, modified=None, size=None):
        self.name = name
        self.modified = modified.strftime("%d-%b-%Y %H:%M") if modified else "-"
        self.size = humanize.naturalsize(size) if size else "-"

    def get_justified_attributes(self, name_length, modified_length, column_padding):
        return {
            "encoded_name": urllib.parse.quote(self.name),
            "name": self.name,
            "name_padding": " ".ljust(name_length - len(self.name) + column_padding),
            "modified": self.modified.ljust(modified_length + column_padding),
            "size": self.size,
        }



# -*- coding: utf-8 -*-
import concurrent.futures
import logging

import boto3
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

from .index import Index
from .item import Item


class BucketDirGenerator:
    def __init__(self, logger=None):
        self.logger = logger or logging.getLogger("bucket_dir")
        self.template_environment = Environment(
            loader=PackageLoader("bucket_dir", "templates"),
            autoescape=select_autoescape(["html"]),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.s3_client = boto3.client("s3")

    def build_indexes(self, contents, exclude_objects):
        indexes = {"/": Index("/")}
        for content in contents:
            key_components = content["Key"].split("/")
            item_name = key_components[-1]
            if len(key_components) > 1:
                path_components = key_components[:-1]
                for depth in range(len(path_components)):
                    path = f"/{'/'.join(path_components[:depth + 1])}/"
                    if depth > 0:
                        parent_path = f"/{'/'.join(path_components[:depth])}/"
                    else:
                        parent_path = "/"
                    if path not in indexes:
                        indexes[path] = Index(path)
                    indexes[parent_path].add_item(Item(f"{path_components[depth]}/"))
                full_path = f"/{'/'.join(path_components)}/"
            else:
                full_path = "/"
            if not exclude_objects:
                exclude_objects = []
            exclude_objects.extend(["", "favicon.ico", "index.html"])
            if item_name not in exclude_objects:
                indexes[full_path].add_item(
                    Item(name=item_name, modified=content["LastModified"], size=content["Size"])
                )
        return indexes

    def generate(
        self, bucket, site_name, exclude_objects=None, single_threaded=False, target_path="/"
    ):
        contents = self.get_bucket_contents(bucket)
        self.logger.info(f"Building indexes for {bucket}.")
        indexes = self.build_indexes(contents, exclude_objects)
        self.logger.info(f"Built {len(indexes)} indexes for {bucket}.")
        if not target_path.startswith("/"):
            target_path = f"/{target_path}"
        descending_indexes = {
            path: index for path, index in indexes.items() if path.startswith(target_path)
        }
        ascending_indexes = {
            path: index for path, index in indexes.items() if target_path.startswith(path)
        }
        target_indexes = {**descending_indexes, **ascending_indexes}
        if single_threaded:
            for path, index in target_indexes.items():
                self.render_and_upload_index_document_to_s3(bucket, site_name, path, index)
        else:
            futures = []
            with concurrent.futures.ThreadPoolExecutor() as executor:
                for path, index in target_indexes.items():
                    futures.append(
                        executor.submit(
                            self.render_and_upload_index_document_to_s3,
                            bucket,
                            site_name,
                            path,
                            index,
                        )
                    )
            for future in futures:
                future.result()
        self.logger.info(f"Finished indexing {bucket} bucket.")

    def get_bucket_contents(self, bucket):
        paginator = self.s3_client.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=bucket)
        self.logger.info(f"Listing objects in the {bucket} bucket.")
        contents = [content for page in page_iterator for content in page["Contents"]]
        self.logger.info(f"Found {len(contents)} objects in the {bucket} bucket.")
        return contents

    def render_and_upload_index_document_to_s3(self, bucket, site_name, path, index):
        self.logger.info(f"Rendering index for {path}.")
        index_document = index.render(
            site_name=site_name, template_environment=self.template_environment
        )
        self.logger.info(f"Uploading index for {path}.")
        key = f"{path[1:]}index.html"
        self.s3_client.put_object(
            Body=index_document.encode(),
            Bucket=bucket,
            CacheControl="max-age=0",
            ContentType="text/html",
            Key=key,
        )



# -*- coding: utf-8 -*-
import logging
import sys

import botocore
import click

from .generator import BucketDirGenerator


def configure_logging(log_level="info") -> None:
    formatter = logging.Formatter(
        fmt="%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger = logging.getLogger("bucket_dir")
    logger.setLevel(log_level.upper())
    logger.addHandler(handler)
    logger.debug("Logging configured")


@click.command()
@click.argument("bucket")
@click.option(
    "--exclude-object",
    help="Exclude objects with this name from indexes. Can be specified multiple times.",
    multiple=True,
)
@click.option(
    "--single-threaded",
    default=False,
    help="Do not using multithreading. Will be slower but less intensive.",
    is_flag=True,
)
@click.option(
    "--target-path",
    default="/",
    help="Only generate indexes relating to a certain folder/object. E.g. '/foo-folder/bar-object'. If omitted, the entire bucket will be indexed.",
)
@click.version_option()
def run_cli(bucket, exclude_object, single_threaded, target_path):
    """Generate directory listings for an S3 BUCKET.

    BUCKET is the name of the bucket to be indexed."""
    configure_logging()
    logger = logging.getLogger("bucket_dir")
    bucket_dir_generator = BucketDirGenerator(logger=logger)
    try:
        bucket_dir_generator.generate(
            bucket=bucket,
            site_name=bucket,
            exclude_objects=list(exclude_object),
            single_threaded=single_threaded,
            target_path=target_path,
        )
    except botocore.exceptions.NoCredentialsError:
        logger.error(
            "Could not discover any AWS credentials. Please supply appropriate credentials."
        )
        sys.exit(1)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bucket_dir']

package_data = \
{'': ['*'], 'bucket_dir': ['templates/*']}

install_requires = \
['Jinja2>=2.11.3,<3.0.0',
 'boto3>=1.17.11,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'humanize>=3.2.0,<4.0.0']

entry_points = \
{'console_scripts': ['bucket-dir = bucket_dir:run_cli']}

setup_kwargs = {
    'name': 'bucket-dir',
    'version': '1.11.0',
    'description': 'Generate directory listings for S3 statically hosted content.',
    'long_description': '# bucket-dir\n\n<a href="https://github.com/hmrc"><img alt="HMRC: Digital" src="https://img.shields.io/badge/HMRC-Digital-FFA500?style=flat&labelColor=000000&logo=gov.uk"></a>\n<a href="https://pypi.org/project/bucket-dir/"><img alt="PyPI" src="https://img.shields.io/pypi/v/bucket-dir"></a>\n<a href="https://pypi.org/project/bucket-dir/"><img alt="Python" src="https://img.shields.io/pypi/pyversions/bucket-dir"></a>\n<a href="https://github.com/hmrc/bucket-dir/blob/master/LICENSE"><img alt="License: Apache 2.0" src="https://img.shields.io/github/license/hmrc/bucket-dir"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n\n**bucket-dir** is a utility for generating a browsable directory tree for an AWS S3 bucket.\n\n!["Sample image"](/docs/sample.png "A sample of bucket-dir output.")\n\nIt was built in order to host Maven and Ivy repositories in S3 and serve them via CloudFront, but it could meet other needs too.\n\n## Installation\n\n```\npip install bucket-dir\n```\n## Usage\n\nRun `bucket-dir` with the name of the bucket you wish to index as a parameter:\n\n```\nbucket-dir foo-bucket\n```\n\nIf you only want to upload indexes for a particular part of the bucket, use `--target-path`. This will generate indexes for folders that lead to the path, and everything under the path:\n\n```\n# These all update the root index, foo-folder\'s index, and everything underneath foo-folder\nbucket-dir foo-bucket --target-path \'/foo-folder/foo-object\'\nbucket-dir foo-bucket --target-path \'/foo-folder/\'\nbucket-dir foo-bucket --target-path \'foo-folder/foo-object\'\nbucket-dir foo-bucket --target-path \'foo-folder/\'\n```\n\nIf you need to exclude objects with certain names from the index use `--exclude-object`. This will hide any objects that match this name. `index.html` objects are ignored for free:\n\n```\nbucket-dir foo-bucket --exclude-object \'error.html\' --exclude-object \'foo-object\'\n```\n\nUse `bucket-dir --help` for all arguments.\n\nBe sure to provide the command with credentials that allow it to perform ListBucket and PutObject calls against the bucket. E.g. with [aws-vault](https://github.com/99designs/aws-vault):\n\n```\naws-vault exec foo-profile -- bucket-dir foo-bucket\n```\n### IAM requirements\n\nThis example demonstrates the most restrictive policy you can apply to the principal (e.g. an IAM user or role) that is going to run `bucket-dir`. Substitute `foo-bucket` for the name of your bucket:\n\n```\n{\n    "Version": "2012-10-17",\n    "Statement": [\n        {\n            "Effect": "Allow",\n            "Action": "s3:ListBucket",\n            "Resource": "arn:aws:s3:::foo-bucket"\n        },\n        {\n            "Effect": "Allow",\n            "Action": "s3:PutObject",\n            "Resource": "arn:aws:s3:::foo-bucket/*"\n        }\n    ]\n}\n```\n\n* `s3:ListBucket` is required for `bucket-dir` to be able to map out the folders and objects that the bucket contains.\n* `s3:PutObject` is required for `bucket-dir` to be able to upload the generated `index.html` documents.\n\n### Example AWS configuration\n\nFor examples on how you can configure an S3 bucket to serve static site content indexed by `bucket-dir`, see:\n\n* [Configuring a public S3 Bucket for use with bucket-dir.](docs/s3_public.md)\n\nExamples of how you can front public and private buckets with CloudFront, and how bucket-dir can be run in a lambda, will be added in due course.\n\n### Using bucket-dir as a library\n\n`bucket-dir` can also be used as a dependency of your own python applications.\n\n```\nfrom bucket_dir import BucketDirGenerator\n\nbucket_dir_generator = BucketDirGenerator()\nbucket_dir_generator.generate(bucket="foo-bucket", site_name="my static site")\n```\n\n### Character support\n\n`bucket-dir` supports objects using any of the _Safe characters_ listed in the S3 [object key naming guidelines](https://docs.aws.amazon.com/AmazonS3/latest/userguide/object-keys.html#object-key-guidelines).\n\nThe exception to the above rule is using forward slashes consecutively (e.g. `my-folder//my-object`). This results in a folder called `/`, which breaks hyperlinks.\n\nUse of characters in the _Characters that might require special handling_ list is currently unsupported but is theoretically ok.\n\nSome characters in _Characters to avoid_ may also work, but you\'re on your own.\n\n## Development\n\nStart with `make init`. This will install prerequisties and set up a poetry managed virtual environment containing all the required runtime and development dependencies.\n\nUnit testing can be performed with `make test`. If you want to run pytest with other options, use `poetry run pytest ...`.\n\nYou can execute the source code directly with `poetry run bucket-dir`.\n\nFinally, you can build with `make build`. This will update dependencies, run security checks and analysis and finally package the code into a wheel and archive.\n\nPublishing can be performed with `make publish`, but this is only intended to run in CI on commit to the main branch. If running locally, you need to have PyPI credentials set as env vars.\n\nFor other rules, see the [Makefile](Makefile).\n\nIf you are a collaborator, feel free to make changes directly to the main branch. Otherwise, please raise a PR. Don\'t forget to bump the version in [pyproject.toml](pyproject.toml).\n\n### Profiling\n\nTo get a performance profile, use:\n\n```\nmake profile\n```\n\nYou must have the `graphviz` library installed.\n\nA `combined.svg` image will be generated in the `prof` directory which you can use to find bottlenecks and potential enhancements.\n\n## License\n\nThis code is open source software licensed under the [Apache 2.0 License]("http://www.apache.org/licenses/LICENSE-2.0.html").\n',
    'author': 'Dave Randall',
    'author_email': '19395688+daveygit2050@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/hmrc/bucket-dir',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
class Index:
    def __init__(self, path):
        self.path = path
        self.items = []

    def add_item(self, item):
        existing_item = [i for i in self.items if i.name == item.name]
        if not existing_item:
            self.items.append(item)

    def render(self, site_name, template_environment):
        template = template_environment.get_template("index.html.j2")
        longest_name_length = max(max([len(item.name) for item in self.items], default=0), 5)
        longest_modified_length = max(
            max([len(item.modified) for item in self.items], default=0), 14
        )
        column_padding = 3
        table_headers = f"{'Name'.ljust(longest_name_length + column_padding)}{'Last modified'.ljust(longest_modified_length + column_padding)}Size"
        index_items = [
            item.get_justified_attributes(
                name_length=longest_name_length,
                modified_length=longest_modified_length,
                column_padding=column_padding,
            )
            for item in self.items
        ]
        return template.render(
            page_name=f"{site_name}{self.path}",
            table_headers=table_headers,
            index_items=index_items,
            not_root=self.path != "/",
        )
